#include <DSOL.hpp>

#include "ExplosionSprite.hpp"


bool ExplosionSprite::isInitialized = false;
uint ExplosionSprite::explosionPaletteIndex = DSOL::Sprites::Palette::NUMBER_OF_PALETTES;
u16* ExplosionSprite::frames[MAX_FRAMES] = {NULL, NULL, NULL, NULL, NULL, NULL, NULL};


/*static*/ void ExplosionSprite::initializeData()
{
    if(!isInitialized)
    {
        using namespace DSOL::Sprites;

        for(uint i = 0; i < MAX_FRAMES; ++i)
        {
            //allocating vram
            frames[i] = oamAllocateGfx(&oamSub, explosionSize, explosionColorFormat);


            //if the program is compiled in debug mode, this will display an error if frames[i] is NULL, and stop the program
            //if its not compiled in debug mode, this will be removed
            sassert(frames[i] != NULL, "failed to allocate vram for explosion frames");


            //calculate the pointer to the right frame. because of how pointer arithmetic works, we have to divide by
                //the size of whatever type explosionTiles is.
            register const uint* framePointer = (explosionTiles + ((i * gfxToByteSize(explosionSize, explosionColorFormat))
                                                                                                    / sizeof(*explosionTiles)));

            //write the right frame to the right vram place
            copyGfxData(framePointer, frames[i], explosionTilesLen / MAX_FRAMES);
        }


        //reserving a palette and writing to it. all frames use the same palette, so we only need 1 palette
        explosionPaletteIndex = Palette::reserveEmptyPalette(false, false);//sub oam and 256 colors
        //another assert to see if the palette index is correctly reserved
        sassert(explosionPaletteIndex != Palette::NUMBER_OF_PALETTES, "no palettes left");
        //copy the palette data
        Palette::writeToPalette(false, false, explosionPaletteIndex, explosionPal, explosionPalLen);

        isInitialized = true;
    }
}



/*static*/ void ExplosionSprite::releaseData()
{
    if(isInitialized)
    {
        //freeing stuff we have used for the second constructor
        DSOL::Sprites::Palette::freePalette(false, false, explosionPaletteIndex); //sub oam and 256 colors
        for(uint i = 0; i < MAX_FRAMES; ++i)
        {
            oamFreeGfx(&oamSub, frames[i]);
            frames[i] = NULL;
        }

        isInitialized = false;
    }
}



ExplosionSprite& ExplosionSprite::setFrame(const uint newFrame /*= 0*/)
{
    sassert(newFrame < MAX_FRAMES, "setFrame() newFrame is out of bounds");

    setGfxPointer(frames[newFrame]);
    currentFrame = newFrame;

    return *this;
}


//the ExplosionSprite constructor
ExplosionSprite::ExplosionSprite(const uint index)
:
//calling Sprites constructor
Sprite(index, explosionSize, explosionColorFormat, getInitFrame(), explosionPaletteIndex,
            false, NULL, MAX_FRAMES, 0),//sub oam, no affine matrix, number of frames and the first frame
//initialising the variables
vblankPerFrame(DSOL::randInt(5,10)), frameVBlanksPassed(0),
vblankPerExplosion(DSOL::randInt(10, 60)), explosionVBlankPassed(DSOL::randInt(0, vblankPerExplosion-1))
{
    using DSOL::randInt;

    setPosition(randInt(0, (SCREEN_WIDTH - getWidth())), randInt(0, (SCREEN_HEIGHT - getHeight())));
    setVisible(false);
    update();
}


void ExplosionSprite::update()
{
    if(explosionVBlankPassed == vblankPerExplosion)
    {
        //the sprite is displaying an explosion
        frameVBlanksPassed++;
        if(frameVBlanksPassed == vblankPerFrame)
        {
            //time for the next frame, or wait if the explosion is over
            register uint newFrame = (getCurrentFrame() + 1);
            if(newFrame == MAX_FRAMES)
            {
                //the explosion is over, so we make the sprite invisible,
                //and wait a number of vblanks
                setVisible(false);
                explosionVBlankPassed = 0;
            }
            else
            {
                //set the next frame
                setFrame(newFrame);
            }

            frameVBlanksPassed = 0;
        }
    }
    else
    {
        //the sprite is waiting between explosions
        explosionVBlankPassed++;
        if(explosionVBlankPassed == vblankPerExplosion)
        {
            //waiting is over, so we set the first frame, set the sprite visible
            //and randomly set the position, frame time(number of vblanks to wait per frame)
            //and wait time(number of vblanks to wait after the explosion finished)
            using DSOL::randInt;
            setFrame(0).setVisible(true);
            setPosition(randInt(0, (SCREEN_WIDTH - getWidth())), randInt(0, (SCREEN_HEIGHT - getHeight())));
            vblankPerFrame = randInt(5,10);
            vblankPerExplosion = randInt(10, 60);
            explosionVBlankPassed = vblankPerExplosion;
        }
    }

    Sprite::update();
}
