#ifndef SPRITEAFFINEMATRIX_HPP_INC
#define SPRITEAFFINEMATRIX_HPP_INC

/** @file   AffineMatrix.hpp
    @brief  contains a class and an enum for managing rotating and scaling data.

    this file contains a class for managing rotating and scaling data for sprites and an enum with some common angles.
*/

#include "IUpdater.hpp"
#include "BooleanArray.hpp"


namespace DSOL
{

namespace Sprites
{


/**
    @brief an enum for the most used angles in the format that libnds uses (32768 degrees).

    the enum contains values for the most used angles in the libnds format. both clockwise as counter clockwise (cc).
*/
enum spriteAngle
{
    Angle_0_degree_cc       = 0,
    Angle_45_degree_cc      = 4096   ,//= 64*64
    Angle_90_degree_cc      = 4096 *2,
    Angle_135_degree_cc     = 4096 *3,
    Angle_180_degree_cc     = 4096 *4,
    Angle_225_degree_cc     = 4096 *5,
    Angle_270_degree_cc     = 4096 *6,
    Angle_315_degree_cc     = 4096 *7,
    Angle_360_degree_cc     = 4096 *8,//= 32768

    Angle_0_degree          = -0,
    Angle_45_degree         = -4096   ,
    Angle_90_degree         = -4096 *2,
    Angle_135_degree        = -4096 *3,
    Angle_180_degree        = -4096 *4,
    Angle_225_degree        = -4096 *5,
    Angle_270_degree        = -4096 *6,
    Angle_315_degree        = -4096 *7,
    Angle_360_degree        = -4096 *8,
};



enum SpriteScale
{
    Scale_25_percent        = (1 << 10),                //(4<<8)
    Scale_50_percent        = (1 << 9),                 //(2<<8)
    Scale_75_percent        = ((1 << 8) + (1 << 7)),
    Scale_100_percent       = (1 << 8),                 //(1<<8)
    Scale_150_percent       = ((1 << 6) + (1 << 7)),
    Scale_200_percent       = (1 << 7),                 //(0.5<<8)
    Scale_400_percent       = (1 << 6)                  //(0.25<<8)
};



//TODO: remove me
//use libnds's degreesToAngle() macro function
//use libnds's angleToDegrees() macro function



/**
 * @brief a class that holds data about rotating and scaling a sprite.
 *
 * sprite(s) can use objects of this class to manage information about scaling and rotating sprites.
 * there can be at most 32 objects (index 0-31) (the hardware wont allow more),
 * but multiple sprites can use the same AffineMatrix object.
 */
class AffineMatrix : implements IUpdater
{
    private:
        //keeps tab on which AffineMatrix objects are being used
        static BooleanArray mainAffineSprites;
        //keeps tab on which AffineMatrix objects are being used
        static BooleanArray subAffineSprites;


    public:

        /** @brief a constant for the number of affine matrixes available per sprite engine */
        static const uint NUMBER_OF_AFFINE_MATRIXES = 32;

        /**
            @brief reserves an empty AffineMatrix index to use.

            this will check if there are AffineMatrix indexes that hasn't been reserved yet,
            and reserves and returns it if there are onreserved AffineMatrix.
            there are 32 AffineMatrix for main oam and 32 for sub engine.

            @param mainOAM      true if the Affine Matrix is for the main oam, false otherwise

            @return the index of an unreserved AffineMatrix index (which has been reserved for you)
                    or NUMBER_OF_AFFINE_MATRIXES if there are no AffineMatrix indexes left (will display an error in debug mode)
        */
        static uint reserveEmptyAffineMatrix(const bool mainOAM = true);


        /**
            @brief frees a AffineMatrix index so other sprites can use it.

            this will free a AffineMatrix index that has been reserved with reserveEmptyAffineMatrix() so that other sprites can use it.

            @param mainOAM                  true if the AffineMatrix index is for the main oam, false otherwise
            @param AffineMatrixIndex  the index to be freed, should be an index gotten with reserveEmptyAffineMatrix()
        */
        static inline void freeAffineMatrix(const uint AffineMatrixIndex, const bool mainOAM = true)
        {
            sassert(AffineMatrixIndex < NUMBER_OF_AFFINE_MATRIXES, "reserveEmptyAffineMatrix() AffineMatrixIndex is out of bounds");

            (mainOAM ? mainAffineSprites : subAffineSprites).set(AffineMatrixIndex, false);
        }



    private:
        //a pointer to which oam should be used
        OamState* oam;

        //the rotation index (0-31)
        uint index;

        //the 4 matrix values
        int hdx, hdy, vdx, vdy;


    public:
        /**
         * @brief the constructor that creates an AffineMatrix object.
         *
         * @param index         the index of the AffineMatrix in oam, must be (0-31).
         * @param useMainOam    true if this AffineMatrix object is for sprites of the main engine, false for the sub engine.
         *                          default is true.
         *
         */
        inline AffineMatrix(const uint index, const bool useMainOam = true)
        : oam(useMainOam? &oamMain: &oamSub), index(index), hdx(1 << 8), hdy(0), vdx(0), vdy(1 << 8)
        {
            sassert(index < NUMBER_OF_AFFINE_MATRIXES, "index AffineMatrix isn't 0-31");
        }


        /**
            @brief the asignment operator, will copy everything exept the index.

            @param source the affine matrix to copy from.

            @return a reference to the object, so you can chain multiple commands.
        */
        inline AffineMatrix& operator= (const AffineMatrix& source)
        {
            if(&source != this)
            {
                //copies everything exept index
                oam = source.oam;
                hdx = source.hdx;
                hdy = source.hdy;
                vdx = source.vdx;
                vdy = source.vdy;
            }

            return *this;
        }


        /**
         * @brief sets the rotation index of the object.
         *
         * @param index the new index of the object.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline AffineMatrix& setIndex(const uint index)
        {
            sassert(index < NUMBER_OF_AFFINE_MATRIXES, "index AffineMatrix setIndex() isn't 0-31");
            this->index = index;

            return *this;
        }


        /**
         * @brief returns the rotation index of the object.
         *
         * @return the rotation index.
         */
        inline const uint getIndex() const
        {
            return index;
        }



        /**
         * @brief sets the AffineMatrix's engine.
         *
         * puts the sprite on the main or the sub engine.
         *
         * @param useMainOam    true if the sprite should be on the main engine, false for the sub engine.
         *                      default is true.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline AffineMatrix& setMainEngine(const bool useMainOam = true)
        {
            oam = (useMainOam ? &oamMain: &oamSub);

            return *this;
        }

        /**
         * @brief returns if the sprite uses the main oam or not.
         *
         * @return true if it uses the main oam, false otherwise
         */
        inline bool getMainEngine() const
        {
            return oam == &oamMain;
        }


        /**
            @brief sets the angle and scale of the object.

            the angle can be one of the spriteAngle enum values,
            a 360 degree value converted with libnds's degreesToAngle() macro function,
            or an 32 bits int if you need precision.
            the scale values are in .8 fixed point format.

            @param angle    the new angle of the object, in the format that libnds uses (32768 degrees), default is 0 (no angle).
            @param scaleX   the value to scale the x ax with.
            @param scaleY   the value to scale the y ax with.

            @return a reference to the object, so you can chain multiple commands.
         */
        AffineMatrix& setAngleScale(const int angle = 0, const int scaleX = (1 << 8), const int scaleY = (1 << 8));



        /**
         * @brief allows you to directly sets the affine transformation matrix
         *
         * with this, you have more freedom to set the matrix, but it might be more difficult to use if
         * you're not used to affine transformation matrix.
         *
         * @param hdx   the 1st matrix value
         * @param hdy   the 2nd matrix value
         * @param vdx   the 3th matrix value
         * @param vdy   the 4th matrix value

            @return a reference to the object, so you can chain multiple commands.
         */
        inline AffineMatrix& setAffineTransformation(const int hdx, const int hdy, const int vdx, const int vdy)
        {
           this->hdx = hdx;
           this->hdy = hdy;
           this->vdx = vdx;
           this->vdy = vdy;

           return *this;
        }


        /**
            @brief returns the 1st value of the affine matrix

            @return the 1st value of the affine matrix
        */
        inline const int gethdx() const
        {
            return hdx;
        }


        /**
            @brief returns the 2nd value of the affine matrix

            @return the 2nd value of the affine matrix
        */
        inline const int gethdy() const
        {
            return hdy;
        }


        /**
            @brief returns the 3th value of the affine matrix

            @return the 3th value of the affine matrix
        */
        inline const int getvdx() const
        {
            return vdx;
        }


        /**
            @brief returns the 4th value of the affine matrix

            @return the 4th value of the affine matrix
        */
        inline const int getvdy() const
        {
            return vdy;
        }



        /*
        /
         * @brief returns the angle of the object.
         *
         * @return the angle in the format that libnds uses (32768 degrees)
         *
        inline int getAngle() const
        {
            return angle;
        }

        /
         * @brief returns the x scale.
         *
         * @return the x scale of the object
         *
        inline int getScaleX() const
        {
            return scaleX;
        }


        /
         * @brief returns the y scale.
         *
         * @return the y scale of the object
         *
        inline int getScaleY() const
        {
            return scaleY;
        }
        */


        /**
         * @brief this will update the AffineMatrix object.
         *
         * this will copy the data to oam so sprites can use it.
         */
        inline virtual void update()
        {
            //the shift by 12 is because of a mistake I made when I recommanded oamAffineTransformation() to libnds.
            //will be fixed.
            oamAffineTransformation(oam, index, hdx<<12, hdy<<12, vdx<<12, vdy<<12);
        }


};//end AffineMatrix class




}// end namespace Sprites

}//end namespace DSOL

#endif // SPRITEAFFINEMATRIX_HPP_INC
