#ifndef BOOLEANARRAY_HPP_INC
#define BOOLEANARRAY_HPP_INC

/** @file   BooleanArray.hpp
    @brief  contains a class for creating a static(with a fixed length) boolean array.
*/


namespace DSOL
{

/**
    @brief a class for creating a static(with a fixed length) boolean array.
*/

class BooleanArray
{
    private:
        //the number of bits in a byte
        static const uint BITS_PER_BYTE = 8;

        //a variable, the bits of the variable will be used as an array of booleans
        uint array;

        //the length of the array
        const uint length;

    public:
        /** @brief the maximum length a BooleanArray can have. */
        static const uint MAX_LENGTH = (sizeof(uint) * BITS_PER_BYTE);

        /**
            @brief creates the array and fills the booleans with false
        */
        inline BooleanArray(uint length) : array(0), length(length)
        {
            sassert(length <= MAX_LENGTH, "error: length of BooleanArray must be equal or lower than MAX_LENGTH");
        }

        /**
            @brief returns the value at the given index

            @param index the index you want to get, starting from 0

            @return the value at the given index
        */
        inline const bool get(const uint index) const
        {
            sassert(index < length, "BooleanArray get() index out of bounds");

            return ((array & BIT(index)) != 0);
        }


        /**
            @brief sets the given value at the given index

            @param index the index of the array that you want to change, starting from 0
            @param value the value that the array at position index should become
        */
        void set(const uint index, const bool value);


        /**
            @brief clears the BooleanArray by setting everything to false
        */
        inline void clear()
        {
            array = 0;
        }

};//end class BooleanArray


}//end namespace DSOL

#endif // BOOLEANARRAY_HPP_INC
