#ifndef LITEVECTOR_HPP_INC
#define LITEVECTOR_HPP_INC

/** @file  LiteVector.hpp
    @brief this file contains a 'lite' version of the std::Vector class.
*/

#include "MemoryFunctions.hpp"



namespace DSOL
{


/**
    @brief a 'lite' version of the std::Vector class.

    this class allows you to use an resizable dynamic array in an oo way without worrying about memory management.
    it doesn't contain exeptions and iterators and stuff like that, which makes it more suitable for devices with low memory.

    it doesn't work with polymorphism (use pointers or references), because it only allocates enough memory for the base type.

    the class uses the types copy constructor and = operator internally, so make sure to implement them correctly.
*/
template <typename T>
class LiteVector
{
    private:
        //the size of the filled in array
        uint size;
        //the size of the allocated buffer
        uint bufferSize;
        //the size that will be added to the buffer if it needs resizing
        const uint incrementSize;
        //a pointer to the beginning of the array
        T* data;


        /* for internal use only
            @brief calls the destructor on the object at index and every object after that.

            @param index    the first object to be destructed, default is 0 (destruct all)
        */
        inline void destruct(uint index = 0)
        {
            if(index > size)
            {
                sassert(false,              "error: index is bigger than size, must be between 0-size");
                index = size;
            }
            sassert(data != NULL,           "memory error: LiteVector data is not initialised");

            if(data != NULL)
            {
                //destruct from last to first
                for(uint i = size; i > index; --i)
                {
                    data[i-1].~T();//calls the destructor
                }

                size = index;
            }
        }


        /* for internal use only
            @brief the 'reconstruct' constructor.

            the reconstruct constructor will destruct the current object and
            copy all the data from source, like the copy constructor.
            this constructor is (almost) exclusively for operator=.
            the 2 LiteVectors will have separate buffers in memory, so try not to copy LiteVectors to much.

            @param source the source to copy from.
            @param unused an unused parameter so the compiler knows the difference
                between the copy constructor and the reconstruct constructor.
        */
        LiteVector(const LiteVector &source, const uint unused)
        : bufferSize(source.bufferSize), incrementSize(source.incrementSize)
        {
            //can be optimised by using operator= where possible

            sassert(this != &source, "error: reconstruct constructor is called with own object as source");

            //call the destructors
            destruct();

            //resize the buffer using the new incrementSize and bufferSize
            reserve(bufferSize);


            //call the constructors
            for(uint i = 0; i < source.size; ++i)
            {
                //use placement new to init the data
                new(&data[i]) T(source.data[i]);
            }

            size = source.size;
        }



    public:
        /**
            @brief the constructor that creates the array.

            @param initSize         the initial size of the allocated memory, default is 16.
            @param incrementSize    the size that will be added to the memory buffer if it automaticly resizes, default is 8.
        */
        LiteVector<T>(const uint initSize = 16, const uint incrementSize = 8)
        : size(0), bufferSize(initSize), incrementSize(incrementSize), data(allocRawMemNoError<T>(initSize))
        {
            sassert(incrementSize != 0, "can't initialise LiteVector, incrementSize can't be 0");
            sassert(data != NULL,       "memory error: can't initialise LiteVector");
        }


        /**
            @brief the Copy constructor, copies the data from 1 LiteVector to the other.

            the 2 LiteVectors will have sepparate buffers in memory, so try not to copy LiteVectors to much.

            @param source the source to copy from.
        */
        LiteVector(const LiteVector &source)
        : size(source.size), bufferSize(source.bufferSize),
          incrementSize(source.incrementSize), data(allocRawMemNoError<T>(source.bufferSize))
        {
            //no need to call destructors
            sassert(data != NULL, "memory error: can't initialise LiteVector");

            //call constructors
            for(uint i = 0; i < size; ++i)
            {
                new(data[i]) T(source.data[i]);
            }
        }


        /**
            @brief the asigment operator, copies the data from 1 LiteVector to the other.

            the 2 LiteVectors will have sepparate buffers in memory, so try not to copy LiteVectors to much.

            @param source the source to copy from.

            @return a reference to the object so you can chain multiple asignments together.
        */
        LiteVector& operator= (const LiteVector &source)
        {
            if(this != &source)
            {
                LiteVector(source, RECONSTRUCT);
            }

            return *this;
        }

        /**
            @brief a comparison operator to compare 2 LiteVectors with each other.

            2 LiteVectors will be considered equal to eachother if the size of the LiteVectors are equal,
            and every item in 1 LiteVector is equal to the items in the other LiteVector.

            @param source the second LiteVector to compare with.
            @return true if both the size and the items of the 2 LiteVectors are equal, false otherwise.
        */
        const bool operator== (const LiteVector &source) const
        {
            if(size != source.size)
            {
                return false;
            }

            for(uint i = 0; i < size; ++i)
            {
                if(data[i] != source.data[i])
                {
                    return false;
                }
            }

            return true;
        }


        /**
            @brief a comparison operator that returns the opposite of operator==.

            @param source the second LiteVector to compare with.
            @return false if both the size and the items of the 2 LiteVectors are equal, true otherwise.
        */
        inline const bool operator!= (const LiteVector &source) const
        {
            return !((*this) == source);
        }


        /**
            @brief gets the item at the given index.

            this will return the item stored at index if index < size,
            and the item stored at 0 if index >= size.
            you can't use this to add items, use set() for that.
                <b>Do NOT set a derived object in the vector this way.
                The vector only has enough memory for the base object (the one specified when creating the vector).</b>

            @param index the index of the array of the item you want, must be between 0 and size-1.

            @return a reference to the item at position given from index,
                    or a reference to the first item if index is out of bounds (will give an error in debug mode).
        */
        //see http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.12
        //for const overloading
        T& get(const uint index) //no const, because you are allowed to modefy the returning items
        {
            sassert(index < size, "error: index is bigger than size, must be between 0-size");
            sassert(data != NULL, "memory error: LiteVector data is not initialised");

            return data[(index < size? index : 0)];
        }


        /**
            @brief returns a const reference of the item at the given index.

            this will return a const reference of the item stored at index if index < size,
            and a const reference of the item stored at 0 if index >= size.
            you can use this function with const LiteVector objects.
            you shouldn't be able to modefy the returned item.

            @param index the index of the array of the item you want, must be between 0 and size-1.

            @return a const reference to the item at position given from index,
                    or a const reference to the first item if index is out of bounds (will give an error in debug mode).
        */
        //see http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.12
        //for const overloading
        const T& get(const uint index) const
        {
            sassert(index < size, "error: index is bigger than size, must be between 0-size");
            sassert(data != NULL, "memory error: LiteVector data is not initialised");

            return data[(index < size? index : 0)];
        }


        /**
            @brief sets an given item at the given index.

            this will overwrite the item at the index with the given item.
            if index is bigger than the size of the array, then the item will be added at the end of the array.

            @param index    the position of the array where you want to store the item
            @param item     a reference to the item you want to store in the array

            @return         true if it succeeded, false if it tried to resize and failed
        */
        const bool set(uint index, const T& item)
        {
            sassert(sizeof(item) == sizeof(T),  "LiteVector set() size of item doesn't match size of items stored in LiteVector");
            if(index > size)
            {
                sassert(false,                  "error: index is bigger than size, must be between 0-size");
                index = size;
            }
            sassert(data != NULL,               "memory error: LiteVector data is not initialised");
            sassert(size <= bufferSize,         "LiteVector array size is bigger than allocated memory");


            //if the item should be added instead of overwriting an other item
            if(index == size)
            {
                //make sure there is an empty space
                if(!reserve(1))
                {
                    return false;
                }

                new(&data[size]) T(item);
                size++;
            }
            else
            {
                //dont destruct the old object, let operator= handle that
                //store the new object
                data[index] = item;
            }

            return true;
        }


        /**
            @brief sets multiple items into the LiteVector.

            this allows you to set multiple items into the LiteVector object.
            the allocated memory will automaticly resize when needed.
            the items can be in any kind of collection that supports the [] operator to
            get acces to the items (standard arrays, LiteVectors, std::Vector, ect).

            @param index        the index of where you want to store the first item, must be 0-size
            @param array        the collection of items you want to store
            @param length       the number of items you want to store
            @param startPoint   the starting index of the collection of items, default is 0.

            @return             true if it worked, false if something went wrong with resizing.
        */
        template<typename T2>
        const bool set(uint index, const T2 array, const uint length, const uint startPoint = 0)
        {
            if (index > size)
            {
                sassert(false, "error: index is bigger than size, must be between 0-size");
                index = size;
            }

            if(index+length > bufferSize)
            {
                if(!reserve((index+length) - size))
                {
                    return false;
                }
            }

            //dont destruct the old objects

            //store the new objects
            for(uint i = 0; i < length; i++)
            {
                sassert(sizeof(array[startPoint + i]) == sizeof(T), "LiteVector set() size of item in array doesn't match size of items stored in LiteVector");

                if((index + i) < size)
                {
                    //overwrite the items using operator=
                    data[index + i] = array[startPoint + i];
                }
                else
                {
                    //initialise the memory using placement new and the copy constructor
                    new(&data[index + i]) T(array[startPoint + i]);
                }
            }

            return true;
        }


        /**
            @brief inserts a given item into the LiteVector.

            @param index    the index at which the item should be inserted.
                            the item at index and everything after it will be moved a spot.
            @param item     the item to insert into the LiteVector.

            @return         true if it succeeded, false if something went wrong with resizing the LiteVector.
        */
        const bool insert(uint index, const T& item)
        {
            sassert(sizeof(item) == sizeof(T),  "LiteVector set() size of item doesn't match size of items stored in LiteVector");
            if(index > size)
            {
                sassert(false,                  "error: index is bigger than size, must be between 0-size");
                index = size;
            }
            sassert(data != NULL,               "memory error: LiteVector data is not initialised");
            sassert(size <= bufferSize,         "LiteVector array size is bigger than allocated memory");

            if(!reserve(1))
            {
                return false;
            }

            //initialise the the last index using placement new and the copy constructor
            new(&data[size]) T(data[size-1]);

            for(uint i = size; i > (index + 1); --i)
            {
                data[i-1] = data[i-2];

            }

            data[index] = item;

            size++;

            return true;
        }




        /**
            @brief allows to acces the items like a normal array.

            acts exactly the same as get().
            you can't use this to add items, use set() for that.
                <b>Do NOT set a derived object in the vector this way. the results are unpredictable.
                The vector only has enough memory for the base object (the one specified when creating the vector).</b>

            @param index    the item you want to get.

            @return         a reference to the item.
        */
        //see http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.12
        //for const overloading
        inline T& operator[](const uint index) //no const, because you are allowed to modify the returning items
        {
            return get(index);
        }


        /**
            @brief allows to acces the items like a normal const array.

            acts exactly the same as the const version of get().
            you shouldn't be able to modefy the returned item.

            @param index    the item you want to get.

            @return         a const reference to the item.
        */
        //see http://www.parashift.com/c++-faq-lite/const-correctness.html#faq-18.12
        //for const overloading
        inline const T& operator[](const uint index) const
        {
            return get(index);
        }


        /**
            @brief removes an item from the array.

            removes the item at index from the array and returns the removed item.
            if the item wasn't the last item of the array, then the last item will be placed at that position
            (thus altering the order).

            @param index the index of the item you want to remove. must be between 0 and size-1.

            @return the removed item
        */
        T remove(uint index)
        {
            if(index > size)
            {
                sassert(false,              "error: index is bigger than size, must be between 0-size");
                index = size;
            }
            sassert(data != NULL,           "memory error: LiteVector data is not initialised");

            //size--;

            if(index == size-1)
            {
                size--;
                //return the last item of the vector
                return data[index];
            }
            else
            {
                register T temp = data[index];

                //move every item 1 place
                for(uint i = index; i < (size - 1); ++i)
                {
                    data[i] = data[i+1];
                }


                //make sure to destroy the last item
                data[size-1].~T();

                size--;
                return temp;
            }
        }


        /**
            @brief adds an item at the end of the vector.

            @param item a reference to the item that will be added to the end of the vector

            @return true if it succeeded, false if it tried to resize and failed
        */
        inline const bool push(const T& item)
        {
            return set(size, item);
        }



        /**
            @brief removes and returns the last item of the vector.

            @return the last item of the vector.
        */
        inline T pop()
        {
            return remove(size);
        }


        /**
            @brief resizes the allocated memory buffer.

            with this, you can manualy resize the buffer. incrementsize will have no effect here.

            @param newBufferSize the new size of the allocated memory buffer.

            @return true if the resizing worked, false if the resizing failed.
        */
        const bool resize(const uint newBufferSize)
        {
            register T* temp = reallocRawMemNoError(data, newBufferSize);

            if(temp == NULL)
            {
                sassert(false, "memory error, resizing LiteVector failed");
                return false;
            }

            //destruct any objects that fall outside of the new buffer
            if(newBufferSize < size)
            {
                destruct(newBufferSize);
            }

            data = temp;
            bufferSize = newBufferSize;

            return true;
        }


        /**
            @brief makes sure that the buffer has extra space for at least the number specified.

            with this function, you can make sure that the buffer has enough space to add items equal to the number specified
            without manually resizing the buffer. usefull if you are going to add alot of items soon.
            this will use incrementsize.

            @param emptySpaceSize the number of empty spaces that should be reserved.

            @return true if there is enough space, false if the resizing failed.
        */
        const bool reserve(const uint emptySpaceSize)
        {
            if((bufferSize - size) > emptySpaceSize)
            {
                //there is already enough space, so no need to resize
                return true;
            }

            //make sure that the new buffer is a multiple of incrementsize
            register uint temp = ((size + emptySpaceSize) / incrementSize);
            return resize((temp + 1) * incrementSize);
        }


        /**
            @brief returns the size of the array.

            @return the size of the array
        */
        inline const uint getSize() const
        {
            return size;
        }


        /**
            @brief returns the size of the allocated memory

            @return the size of the allocated memory
        */
        inline const uint getBufferSize() const
        {
            return bufferSize;
        }



        /**
            @brief clears the LiteVector of its contents.
        */
        inline void clear()
        {
            destruct();
        }



        /**
            @brief the destructor that deallocates the memory
        */
        inline ~LiteVector<T>()
        {
            destruct();
            deallocRawMemNoError<T>(data);
        }

};//end LiteVector class

}//end namespace DSOL

#endif // LITEVECTOR_HPP_INC
