#ifndef MEMORYFUNCTIONS_HPP_INC
#define MEMORYFUNCTIONS_HPP_INC

/** @file MemoryFunctions.hpp
    @brief contains some template wrappers to make malloc easier and better to work with.
*/


//I use malloc.h, which is already included in nds.h
//#include <cmalloc>

#include <new>//for placement new


//=======================================================================//
//http://forum.gbadev.org/viewtopic.php?t=14438&highlight=getmemfree     //
//=======================================================================//
//needed for the memory information
#include <unistd.h>
//#include <malloc.h>

extern u8 __end__[];        // end of static code and data
extern u8 __eheap_end[];    // farthest point to which the heap will grow



namespace DSOL
{

/**
    @brief a template function to allocate raw memory.

    this function will allocate memory for the type and number specified,
    but it wont call constructors of check whether NULL was returned.

    @param length how many memory it should allocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
inline T* allocRawMemNoError(const uint length = 1)
{
    return static_cast<T*>(malloc(sizeof(T) * length));
}


/**
    @brief a template function to reallocate raw memory.

    this function will reallocate memory for the type and number specified,
    but it wont call constructors/destructors of check whether NULL was returned.

    @param oldMemPointer a pointer to the previously allocated memory.
    @param length how many memory it should reallocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
inline T* reallocRawMemNoError(T* oldMemPointer, const uint length = 1)
{
    //realloc should handle NULL pointers itself, so no need to check oldMemPointer for NULL
    return static_cast<T*>(realloc(oldMemPointer, sizeof(T) * length));
}


/**
    @brief a template function to deallocate raw memory.

    this function will deallocate memory and set the pointer to NULL, but it wont call destructors.

    @param oldMemPointer a referrence to a pointer to the previously allocated memory.
*/
template <typename T>
inline void deallocRawMemNoError(T*& oldMemPointer)//a referrence to a pointer (I hope)
{
    if(oldMemPointer != NULL)
    {
        free(oldMemPointer);
        oldMemPointer = NULL;
    }
}




/**
    @brief a template function to allocate raw memory.

    this function will allocate memory for the type and number specified, but it wont call constructors.
    it will check whether NULL was returned and display an error with sassert() if it does.

    @param length how many memory it should allocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory (will sassert() in debug mode).
*/
template <typename T>
inline T* allocRawMem(const uint length = 1)
{
    register T* temp = allocRawMemNoError<T>(length);
    sassert(temp != NULL, "memory error: failed to allocate memory");
    return temp;
}


/**
    @brief a template function to reallocate raw memory.

    this function will reallocate memory for the type and number specified, but it wont call constructors/destructors.
    it will check if oldMemPointer is NULL and whether NULL was returned and display an error with sassert() if it does.

    @param oldMemPointer a pointer to the previously allocated memory.
    @param length how many memory it should reallocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory (will sassert() in debug mode).
*/
template <typename T>
inline T* reallocRawMem(T* oldMemPointer, const uint length = 1)
{
    sassert(oldMemPointer != NULL, "reallocRawMem() oldMemPointer is NULL");
    register T* temp = reallocRawMemNoError<T>(oldMemPointer, length);
    sassert(temp != NULL, "memory error: failed to reallocate memory");
    return temp;
}


/**
    @brief a template function to deallocate raw memory.

    this function will deallocate memory and set the pointer to NULL, but it wont call destructors.
    it will check whether oldMemPointer is NULL and and display an error with sassert() if it is.

    @param oldMemPointer a referrence to a pointer to the previously allocated memory.
*/
template <typename T>
inline void deallocRawMem(T*& oldMemPointer)//a referrence to a pointer (I hope)
{
    sassert(oldMemPointer != NULL, "deallocRawMemNoError() oldMemPointer is NULL");
    deallocRawMemNoError<T>(oldMemPointer);
}



/**
    @brief a template function to allocate memory.

    this function will allocate memory for the type and number specified
    and calls the copy constructor using the specified value.
    however it won't check whether NULL was returned or not.

    @param value    the object that should be used when calling the copy constructor, default is the default constructor.
    @param length   how many memory it should allocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
T* allocMemNoError(const T value = T(), const uint length = 1)
{
    register T* temp = allocRawMemNoError<T>(length);

    if(temp != NULL)
    {
        for(uint i = 0; i < length; ++i)
        {
            new(temp + i) T(value);//using placement new to call the copy constructor which copies from value
        }
    }

    return temp;
}


/**
    @brief a template function to reallocate memory.

    this function will reallocate memory for the type and number specified and it will call constructors/destructors.
    however it won't check whether NULL was returned or not.

    @param oldMemPointer    a pointer to the previously allocated memory.
    @param oldLength        the length of the previously allocated memory, needed to correctly call constructors/destructors.
    @param length           how many memory it should reallocate (how long the buffer should be), default is 1.
    @param value            the object that should be used when calling the copy constructor, default is the default constructor.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
T* reallocMemNoError(T* oldMemPointer, const uint oldLength, const T value = T(), const uint length = 1)
{
    if(length == oldLength)
    {
        return oldMemPointer;
    }
    else if(length < oldLength)
    {
        //destruct from last to first
        for(uint i = oldLength; i > length; --i)
        {
            oldMemPointer[i-1].~T();
        }
    }

    register T* temp = reallocRawMemNoError<T>(oldMemPointer, length);

    if(temp != NULL && length > oldLength)
    {
        for(uint i = length; i < oldLength; ++i)
        {
            new(temp + i) T(value);//using placement new to call the copy constructor which copies from value
        }
    }

    return temp;
}


/**
    @brief a template function to deallocate memory.

    this function will deallocate memory, set the pointer to NULL and call the destructors.

    @param oldMemPointer    a referrence to a pointer to the previously allocated memory.
    @param oldLength        the length of the previously allocated memory, needed to correctly call destructors.
*/
template <typename T>
void deallocMemNoError(T*& oldMemPointer, const uint oldLength)//a referrence to a pointer (I hope)
{
    if(oldMemPointer != NULL)
    {
        //destruct from last to first
        for(uint i = oldLength; i > 0; --i)
        {
            oldMemPointer[i-1].~T();
        }

        deallocRawMemNoError<T>(oldMemPointer);
    }
}



/**
    @brief a template function to allocate memory.

    this function will allocate memory for the type and number specified and calls the copy constructor using the specified value.
    it will check whether NULL was returned and display an error with sassert() if it does.

    @param value    the object that should be used when calling the copy constructor, default is the default constructor.
    @param length   how many memory it should allocate (how long the buffer should be), default is 1.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
inline T* allocMem(const T value = T(), const uint length = 1)
{
    register T* temp = allocMemNoError<T>(value, length);
    sassert(temp != NULL, "memory error: failed to allocate memory");
    return temp;
}


/**
    @brief a template function to reallocate memory.

    this function will reallocate memory for the type and number specified and it will call constructors/destructors.
    it will check if oldMemPointer is NULL and whether NULL was returned and display an error with sassert() if it does.

    @param oldMemPointer    a pointer to the previously allocated memory.
    @param oldLength        the length of the previously allocated memory, needed to correctly call constructors/destructors.
    @param length           how many memory it should reallocate (how long the buffer should be), default is 1.
    @param value            the object that should be used when calling the copy constructor, default is the default constructor.

    @return a pointer to the allocated memory, or NULL if it failed to allocate the memory.
*/
template <typename T>
inline T* reallocMem(T* oldMemPointer, const uint oldLength, const uint length = 1, const T value = T())
{
    sassert(oldMemPointer != NULL, "reallocMem() oldMemPointer is NULL");
    register T* temp = reallocMemNoError<T>(oldMemPointer, oldLength, length, value);
    sassert(temp != NULL, "memory error: failed to reallocate memory");
    return temp;
}


/**
    @brief a template function to deallocate memory.

    this function will deallocate memory, set the pointer to NULL and call the destructors.
    it will check whether oldMemPointer is NULL and and display an error with sassert() if it is.

    @param oldMemPointer    a referrence to a pointer to the previously allocated memory.
    @param oldLength        the length of the previously allocated memory, needed to correctly call destructors.
*/
template <typename T>
inline void deallocMem(T*& oldMemPointer, const uint oldLength = 1)//a referrence to a pointer (I hope)
{
    sassert(oldMemPointer != NULL, "deallocMem() oldMemPointer is NULL");

    deallocMemNoError<T>(oldMemPointer, oldLength);
}







//=======================================================================//
//http://forum.gbadev.org/viewtopic.php?t=14438&highlight=getmemfree     //
//=======================================================================//


static inline u8* getHeapStart()
{
   return __end__;
}


static inline u8* getHeapEnd()
{
   return (u8*)sbrk(0);
}


static inline u8* getHeapLimit()
{
   return __eheap_end;
}


static inline int getMemUsed()
{
   struct mallinfo mi = mallinfo();
   return mi.uordblks;
}


static inline int getMemFree()
{
   struct mallinfo mi = mallinfo();
   return mi.fordblks + (getHeapLimit() - getHeapEnd());
}



}//end namespace DSOL


#endif // MEMORYFUNCTIONS_HPP_INC
