#ifndef PALETTE_HPP_INC
#define PALETTE_HPP_INC

#include "standardDefines.hpp"
#include "BooleanArray.hpp"

/** @file Palette.hpp
    @brief this file contains a class for handling which sprite palettes are in use.
*/

namespace DSOL
{

//temp fix until it is included with libnds
#define VRAM_I_EXT_PALETTE ((_ext_palette *)VRAM_I)


namespace Sprites
{


/**
    @brief a static class that manages empty sprite palettes.
*/
class Palette
{
    public:
        /** @brief a constant for the number of palettes there are for each mode. */
        static const uint NUMBER_OF_PALETTES = 16;


    private:
        //keeps tab on which palette index are already taken(true), for both 16 and 256 color palettes (in that order)
        static BooleanArray mainPalette;

        //keeps tab on which palette index are already taken(true), for both 16 and 256 color palettes (in that order)
        static BooleanArray subPalette;

        //an constant to multiply a palette index with to get the right palette ofset
        static const uint PALETTE_INDEX_MULTIPLIER = 16;

    public:

        /**
            @brief reserves an empty palette to use.

            this will check if there are palettes that hasn't been reserved yet,
            and reserves and returns it if there are onreserved palettes.
            there are 16 palettes for 16 color sprites and 16 palettes for 256 color sprites

            @param mainOAM      true if the palette is for the main oam, false otherwise
            @param palette16c   true if the palette needs to be for 16 color sprites, false for 256 color sprites

            @return the index of an unreserved palette (which has been reserved for you)
                    or NUMBER_OF_PALETTES if there are no palettes left (will display an error in debug mode)
        */
        static uint reserveEmptyPalette(const bool mainOAM, const bool palette16c);



        /**
            @brief frees a palette so other sprites can use it.

            this will free a palette that has been reserved with reserveEmptyPalette() so that other sprites can use it.

            @param mainOAM      true if the palette is for the main oam, false otherwise.
            @param palette16c   true if the palette needs to be for 16 color sprites, false for 256 color sprites.
            @param paletteIndex the index to be freed, should be an index gotten with reserveEmptyPalette().
        */
        static void freePalette(const bool mainOAM, const bool palette16c, const uint paletteIndex);



        /**
            @brief writes the palette data for 16 colored sprites into the right palette.

            @param mainOAM          true if the palette data will be used for main OAM, false for sub OAM.
            @param paletteIndex     the index to the palette, preferably obtained with using reserveEmptyPalette().
            @param palettePointer   a pointer to the palette data.
            @param paletteLength    the length of the data in bytes, must be 0 - 16*sizeof(u16).
        */
        static void writeToPalette16c(const bool mainOAM, const uint paletteIndex, const u16* palettePointer,
                                        const uint paletteLength);



        /**
            @brief writes the palette data for 256 colored sprites into the right palette.

            @param mainOAM          true if the palette data will be used for main OAM, false for sub OAM.
            @param paletteIndex     the index to the palette, preferably obtained with using reserveEmptyPalette().
            @param palettePointer   a pointer to the palette data.
            @param paletteLength    the length of the data in bytes, must be 0 - 256*sizeof(u16).
        */
        static void writeToPalette256c(const bool mainOAM, const uint paletteIndex, const u16* palettePointer,
                                        const uint paletteLength);




        /**
            @brief copies the palette data for sprites into the right palette.

            @param mainOAM          true if the palette data will be used for main OAM, false for sub OAM.
            @param palette16c       true if the palette is for 16 color sprites, false for 256 color sprites.
            @param paletteIndex     the index to the palette, preferably obtained with using reserveEmptyPalette().
            @param palettePointer   a pointer to the palette data.
            @param paletteLength    the length of the data in bytes.
        */
        static inline void writeToPalette(const bool mainOAM, const bool palette16c, const uint paletteIndex,
                                    const u16* palettePointer, const uint paletteLength)
        {
            if(palette16c)
            {
                writeToPalette16c(mainOAM, paletteIndex, palettePointer, paletteLength);
            }
            else
            {
                writeToPalette256c(mainOAM, paletteIndex, palettePointer, paletteLength);
            }
        }


        /**
            @brief copies 16 color palette A to palette B.

            @param paletteAmainOAM  true if palette A is for main OAM, false for sub OAM.
            @param paletteAIndex    the index of palette A.
            @param paletteBmainOAM  true if palette B is for main OAM, false for sub OAM.
            @param paletteBIndex    the index of palette B.
        */
        static void copyPalette16c(const bool paletteAmainOAM, const uint paletteAIndex, const bool paletteBmainOAM,
                                    const uint paletteBIndex);



        /**
            @brief copies 16 color palette A to palette B.

            @param paletteAmainOAM  true if palette A is for main OAM, false for sub OAM.
            @param paletteAIndex    the index of palette A.
            @param paletteBmainOAM  true if palette B is for main OAM, false for sub OAM.
            @param paletteBIndex    the index of palette B.
        */
        static void copyPalette256c(const bool paletteAmainOAM, const uint paletteAIndex, const bool paletteBmainOAM,
                                    const uint paletteBIndex);



        /**
            @brief copies palette A to palette B.

            both palettes must have the same color length(16 or 256).

            @param palettes16c      true if both palettes are for 16 color sprites, false for 256 color.
            @param paletteAmainOAM  true if palette A is for main OAM, false for sub OAM.
            @param paletteAIndex    the index of palette A.
            @param paletteBmainOAM  true if palette B is for main OAM, false for sub OAM.
            @param paletteBIndex    the index of palette B.
        */
        static inline void copyPalette(const bool palettes16c, const bool paletteAmainOAM, const uint paletteAIndex,
                                        const bool paletteBmainOAM, const uint paletteBIndex)
        {
            if(palettes16c)
            {
                copyPalette16c(paletteAmainOAM, paletteAIndex, paletteBmainOAM, paletteBIndex);
            }
            else
            {
                copyPalette256c(paletteAmainOAM, paletteAIndex, paletteBmainOAM, paletteBIndex);
            }
        }



};//end class Palette



}//end namespace Sprites

}//end namespace DSOL

#endif // PALETTE_HPP_INC
