#ifndef SPRITE_HPP_INC
#define SPRITE_HPP_INC

/** @file   Sprite.hpp
    @brief  this file contains a Sprite class for handling sprites on the DS.
*/

#include "standardDefines.hpp"
#include "MemoryFunctions.hpp"
#include "IUpdater.hpp"

#include "AffineMatrix.hpp"
#include "Palette.hpp"



namespace DSOL
{

namespace Sprites
{


/**
    @brief returns the width for a given SpriteSize enum value.

    @param size the SpriteSize enum value.

    @return the width, 0 if the SpriteSize wasn't recognized (shouldn't happen tho).
*/
const uint spriteSizeToWidth(const SpriteSize size);


/**
    @brief returns the height for a given SpriteSize enum value.

    @param size the SpriteSize enum value.

    @return the height, 0 if the SpriteSize wasn't recognized (shouldn't happen tho).
*/
const uint spriteSizeToHeight(const SpriteSize size);


/**
    @brief returns the number of pixels a SpriteSize uses.

    @param size the SpriteSize enum value.

    @return the number of pixels (height * width).
*/
static inline const uint spriteSizeToPixel(const SpriteSize size)
{
    return (spriteSizeToHeight(size) * spriteSizeToWidth(size));
}


/**
    @brief returns the size in bytes of a given size and colorFormat.

    @param size         the SpriteSize enum value.
    @param colorFormat  the SpriteColorFormat enum value.

    @return the size in bytes.
*/
const uint gfxToByteSize(const SpriteSize size, const SpriteColorFormat colorFormat);


/**
    @brief calculates the number of frames that a sprite has.

    @param gfxLength    the length of the graphics data.
    @param size         the size of the sprite, 1 of libnds enum values.
    @param colorFormat    the number of bits a pixel has, 1 of libnds enum values.

    @return the number of frames that an image has.
*/
static inline const uint gfxToFrameCount(const uint gfxLength, const SpriteSize size, const SpriteColorFormat colorFormat)
{
    return gfxLength / gfxToByteSize(size, colorFormat);
}


/**
    @brief a wrapper for copying gfx data.

    @param sourcePointer    a pointer to the gfx data.
    @param gfxPointer       a pointer to the destination in vram.
    @param dataLenght       the length of the data.
*/
static inline void copyGfxData(const u32* sourcePointer, u16* gfxPointer, const uint dataLength)
{
    dmaCopyHalfWords(DMA_PRIORITY, sourcePointer, gfxPointer, dataLength);
}



/**
    @brief a class for working with ds sprites.

    this class has 2 constructors. the first one will allocate vram, palette index, ect itself, which makes it very easy to work with.
    the second constructor requires a bit more work (like allocating vram yourself), but it is more flexible
    (like allowing you to choose how to allocate vram).
    be very carefull with asigning sprites and vram ownership, so you won't get sprites with invalid vram.
 */
class Sprite : implements IUpdater
{
    public:
        /** @brief a constant for the number of sprites thats available per engine. */
        static const uint NUMBER_OF_SPRITES = 128;
        /** @brief a constant for the number of priorities that exists. */
        static const uint NUMBER_OF_PRIORITIES = 4;


    private:
        //a pointer to which oamPointer should be used
        OamState* oamPointer;

        //the oam index, should be between 0-127
        uint index;

        //the priority of the sprite (0-3)
        //the lower the priority, the later it is drawed, the more on top it is
        uint priority;

        //a pointer to the graphics stored in vram
        u16* gfxPointer;

        //the x position of the sprite
        int xPos;

        //the y position of the sprite
        int yPos;


        //the number of bits that a pixel uses, as an libnds enum variable
        SpriteColorFormat colorFormat;

        //the size of the sprite as an libnds enum variable
        SpriteSize size;

        //the palette index if multiple palettes or the alpha value if bmp sprite
        int paletteIndexOrAlphaValue;


        //a pointer to a SpriteRotateScale object that holds information about rotating and scaling
        //seeing as more than 1 sprite can point to the same SpriteRotateScale object,
        //this class only has a pointer so if this object gets destroyed, the SpriteRotateScale object still exist
        AffineMatrix *affineMatrixPointer;


        //is the sprite visible or hidden
        bool visible;

        //flip the sprite vertical?
        bool flipVertical;

        //flip the sprite horizontal?
        bool flipHorizontal;

        //apply mosaic
        bool mosaic;

        //double the size when rotating?
        bool doubleSizeOnRotation;


        //===============================================================================

        //should it release the allocated graphics in the destructor
        mutable bool ownsGraphics;
        //should it release the palette in the destructor
        mutable bool ownsPalette;
        //should it release the RotateScale object in the destructor
        mutable bool ownsAffineMatrix;


        //a pointer to the image graphics data
        const u32* dataPointer;

        //the length of the data
        const uint dataLength;

        //a pointer to the image palette data
        const u16* palettePointer;

        //the length of the palette
        const uint paletteLength;



    protected:

        /**
            @brief the number of frames the image has.

            a protected member that you can use in your derived classes.
            this is used when using the first constructor, so make sure not to change it unless you use the second constructor.
        */
        uint maxNumberOfFrames;

        /**
            @brief the frame the image currently is on.

            a protected member that you can use in your derived classes.
            this is used when using the first constructor, so make sure not to change it unless you use the second constructor.
        */
        uint currentFrame;




        /**
            @brief releases the graphics pointer if the sprite owns one.
        */
        void releaseGFX();


        /**
            @brief releases the palette if the sprite owns one.
        */
        void releasePalette();


        /**
            @brief releases the affine matrix if the sprite owns one.
        */
        void releaseAffineMatrix();


        /**
            @brief an alternative version of the second sprite constuctor.

            with this protected constructor, you can set the protected variables maxNumberOfFrames and currentFrame.

            @param index                    the index into the oam (0-127), must be unique for every sprite in an engine.
            @param size                     the size of the sprite, as an SpriteSize enum defined in libnds.
            @param colorFormat              the number of bits that a pixel uses, as an libnds enum variable.
            @param gfxPointer               a pointer to the graphics data.
            @param paletteIndexOrAlphaValue the palette index, or the alpha value if ColorFormat is bmp.
            @param useMainOam               true if the sprite should be on the main engine, false for the sub engine.
            @param affineMatrixPointer      a pointer to a AffineMatrix class that holds the Matrix.
            @param maxNumberOfFrames        a value for maxNumberOfFrames. not used by this constructor, so usable by any derived class.
            @param currentFrame             a value for currentFrame. not used by this constructor, so usable by any derived class.
         */
        inline Sprite(const uint index, const SpriteSize size, const SpriteColorFormat colorFormat,
            u16* gfxPointer, const uint paletteIndexOrAlphaValue, const bool useMainOam,
            AffineMatrix *affineMatrixPointer, const uint maxNumberOfFrames, const uint currentFrame)
            :
            oamPointer(useMainOam? &oamMain: &oamSub), index(index), priority(0), gfxPointer(gfxPointer),
            xPos(0), yPos(0), colorFormat(colorFormat), size(size), paletteIndexOrAlphaValue(paletteIndexOrAlphaValue),
            affineMatrixPointer(affineMatrixPointer), visible(true), flipVertical(false), flipHorizontal(false),
            mosaic(false), doubleSizeOnRotation(false),

            ownsGraphics(false), ownsPalette(false), ownsAffineMatrix(false),
            dataPointer(NULL), dataLength(0), palettePointer(NULL), paletteLength(0),
            maxNumberOfFrames(maxNumberOfFrames), currentFrame(currentFrame)
        {
            sassert(gfxPointer != NULL, "gfxPointer of sprite is NULL");
            sassert(index < NUMBER_OF_SPRITES, "index Sprite is not in 0-127");
        }



        /**
            @brief the 'reconstruct' constructor, recreates an already created object.

            the reconstruct constructor will destruct the current object and
            copy all the data from source, like the copy constructor.
            you can specify if the sprite should allocate new vram, palette, ect or should use the sources vram, palette ect.
            this constructor is (almost) exclusively for operator=.

            @param source       the source to copy from.
            @param allocateNew  true if the sprite should allocate new vram ect, false if it should just share the sources vram, ect.
            @param takeOwnership true if the sprite should take ownership, false for not. ignored if allocateNew is true.
            @param unused       an unused parameter so the compiler knows the difference between the copy constructor and
                                the reconstruct constructor. unneeded here, but kept for consistency.

        */
        Sprite(const Sprite& source, const bool allocateNew, const bool takeOwnership, const uint unused = 0);




    public:
        /**
            @brief The first sprite constructor, which will allocate everything itself (vram, palette, affine matrix, ect).

            Sprites constructor 1, will allocate vram, palette ect and also copy graphics, palette ect.
            This is the easiest way to create a sprite, and should be enough most of the time.
            The constuctor will create a sprite with the given information, but it will not update it.
            After creating the sprite, you should change the information you want,
            and then update the sprite yourself with the update function.
            This system also allows you to set the frame of the sprite, because the sprite knows all the frames.

            @param index            the index into the oam (0-127), must be unique for every sprite in an engine.
            @param dataPointer      a pointer to the graphics data.
            @param dataLength       the length of the graphics data.
            @param palettePointer   a pointer to the palette data. if NULL, palette wont be copied,
                                        and user has to specify which palette to use with paletteLength.
                                        ignored if ColorFormat is bmp.
            @param paletteLength    the lenght of the palette data, or the palette index if palettePointer is NULL,
                                        or the alpha value if ColorFormat is bmp.
            @param size             the size of the sprite, as an libnds SpriteSize enum value.
            @param colorFormat      the number of bits that a pixel uses, as an libnds SpriteColorFormat enum value.
            @param useMainOam       true if the sprite is for the main engine, false otherwise.          default is true.
            @param useAffineMatrix  true if the sprite needs an affine matrix, false otherwise.          default is false.
        */
        Sprite( const uint index,
                const u32* dataPointer, const uint dataLength,
                const u16* palettePointer, const uint paletteLength,
                const SpriteSize size, const SpriteColorFormat colorFormat,
                const bool useMainOam = true,
                const bool useAffineMatrix = false);


        /**
            @brief The second sprite constuctor, which doesn't allocate vram or copies data.

            Sprites constructor 2, doesn't allocate vram, palette ect and also doesn't copy graphics, palette ect.
            You have to do alot more this way, but you will have more freedom doing it.
            Do make sure that it doesn't conflict with the library system.
            The constuctor will create a sprite with the given information, but it will not update it.
            After creating the sprite, you should change the information you want,
            and then update the sprite yourself with the update function.
            This system doesn't allows you to set the frame of the sprite, because the sprite doesn't knows all the frames.
            the maximum number of frames is always 1 and the current frame is always 0.

            @param index                    the index into the oam (0-127), must be unique for every sprite in an engine.
            @param size                     the size of the sprite, as an SpriteSize enum defined in libnds.
            @param colorFormat              the number of bits that a pixel uses, as an libnds enum variable.
            @param gfxPointer               a pointer to the graphics data.
            @param paletteIndexOrAlphaValue the palette index, or the alpha value if ColorFormat is bmp.
            @param useMainOam               true if the sprite should be on the main engine, false for the sub engine.  default is true.
            @param affineMatrixPointer      a pointer to an AffineMatrix object that holds the Matrix.                  default is NULL.

         */
        inline Sprite(const uint index, const SpriteSize size, const SpriteColorFormat colorFormat,
            u16* gfxPointer, const uint paletteIndexOrAlphaValue, const bool useMainOam = true,
            AffineMatrix *affineMatrixPointer = NULL)
            :
            oamPointer(useMainOam? &oamMain: &oamSub), index(index), priority(0), gfxPointer(gfxPointer),
            xPos(0), yPos(0), colorFormat(colorFormat), size(size), paletteIndexOrAlphaValue(paletteIndexOrAlphaValue),
            affineMatrixPointer(affineMatrixPointer), visible(true), flipVertical(false), flipHorizontal(false),
            mosaic(false), doubleSizeOnRotation(false),

            ownsGraphics(false), ownsPalette(false), ownsAffineMatrix(false),
            dataPointer(NULL), dataLength(0), palettePointer(NULL), paletteLength(0),
            maxNumberOfFrames(1), currentFrame(0)
        {
            sassert(gfxPointer != NULL, "gfxPointer of sprite is NULL");
            sassert(index < NUMBER_OF_SPRITES, "index Sprite is not in 0-127");
        }


        /**
            @brief the Copy constructor, copies the data from 1 Sprite to the other.

            the copy constructor won't allocate vram, palette ect itself, but uses the sources stuff.
            it will also take ownership of the vram ect by default (so you can use it correctly with std::vector or LiteVector).

            @param source           the source sprite to copy from.
            @param takeOwnership    true if the sprite should take ownership of the vram ect,
                                    false if the source should keep ownership. default is true.
        */
        Sprite(const Sprite &source, const bool takeOwnership = true);


        /**
            @brief the asigment operator, copies the data from 1 Sprite to the other.

            the copy constructor won't allocate vram, palette ect itself, but uses the sources stuff.
            it will also take ownership of the vram ect (so you can use it correctly with std::vector or LiteVector).

            @param source the source to copy from.

            @return a reference to the object, so you can chain multiple commands (including asignments).
        */
        inline Sprite& operator= (const Sprite &source)
        {
            if(&source != this)
            {
                Sprite(source, false, true, RECONSTRUCT);
            }

            return *this;
        }



        /**
            @brief allocates vram for the sprite, if it doesn't already owns vram.

            with this function, you can let the sprite allocate vram, just like the first constructor would have done.
            however, it doesn't copy data to it, so you should copy the data yourself, or else there will be garbage on the screen.

            you can always use this function, even if you used the second constructor. however, if you're planning to use this,
            then you should try to just let the first constructor handle it.

            @return a reference to the object, so you can chain multiple commands.
        */
        Sprite& allocateGraphics();


        /**
            @brief allocates a palette index, if the sprite doesn't own a palette index already.

            with this function, you can let the sprite allocate a palette index, just like the first constructor would have done.
            however, it won't copy data to the palette, so you should copy the data to the palette yourself,
            or else the sprite will have random colors.

            you can always use this function, even if you used the second constructor. however, if you're planning to use this,
            then you should try to just let the first constructor handle it.

            @return a reference to the object, so you can chain multiple commands.
        */
        Sprite& allocatePalette();


        /**
            @brief allocates a new affine matrix, if the sprite doesn't own a matrix already.

            with this function, you can let the sprite allocate a matrix, just like the sprite would have done
            if you passed true to the affine matrix parameter of the first constructor.
            you can always use this function, even if you used the second constructor.

            @return a reference to the object, so you can chain multiple commands.
        */
        Sprite& allocateAffineMatrix();



        /**
            @brief writes data directly to the Sprites vram.

            this can be used with the second constructor to overwrite the sprite, or the first constructor if
            you didn't provide the data with the constructor (it is recommanded to pass all the data with the first constructor).

            @param dataPointer a pointer to the graphics data.
            @param dataLength the length of the data in bytes.

            @return a reference to the object, so you can chain multiple commands.
        */
        inline Sprite& writeGraphics(const u32* dataPointer, const uint dataLength)
        {
            sassert(dataLength <= gfxToByteSize(size, colorFormat), "error: writeGraphics() dataLength is to big");

            copyGfxData(dataPointer, gfxPointer, dataLength);

            return *this;
        }



        /**
            @brief writes to the palette of the Sprite.

            this can be used to overwrite the palette.

            @param dataPointer  a pointer to the palette data.
            @param dataLength   the length of the data in bytes.

            @return a reference to the object, so you can chain multiple commands.
        */
        inline Sprite& writePalette(const u16* dataPointer, const uint dataLength)
        {
            Palette::writeToPalette(oamPointer == &oamMain, colorFormat == SpriteColorFormat_16Color, paletteIndexOrAlphaValue,
                            dataPointer, dataLength);

            return *this;
        }




        /**
         * @brief sets the index of the sprite.
         *
         * @param index the new index of the sprite.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setIndex(const uint index)
        {
            sassert(index < NUMBER_OF_SPRITES, "index Sprite is not in 0-127");
            this->index = index;

            return *this;
        }


        /**
         * @brief returns the index of the sprite.
         *
         * @return the index of the sprite.
         */
        inline const int getIndex() const
        {
            return index;
        }



        /*// ***needs work***
         * @brief sets the sprite's engine.
         *
         * puts the sprite on the main or the sub engine.
         *
         * @param gfxPointer    a pointer to the new graphics data,
         *                      seeing as graphics data for main and sub engine are stored in different places.
         * @param useMainOam    true if the sprite should be on the main engine, false for the sub engine.
         *                      default is true.
         *
        inline void setMainEngine(u16* gfxPointer, const bool useMainOam = true)
        {
            sassert(gfxPointer != NULL, "gfxPointer of sprite is NULL");
            this->gfxPointer = gfxPointer;
            oamPointer = (useMainOam == true ? &oamMain: &oamSub);
        }*/


        /**
         * @brief returns if the sprite uses the main oam or not.
         *
         * @return true if it uses the main oam, false otherwise.
         */
        inline const bool isOnMainEngine() const
        {
            return (oamPointer == &oamMain);
        }


        /**
         * @brief returns the size of the sprite.
         *
         * @return the size of the sprite, as an libnds enum value.
         */
        inline const SpriteSize getSize() const
        {
            return size;
        }


        /**
            @brief returns the width of the sprite.

            @return the width of the sprite.
        */
        inline const uint getWidth() const
        {
            return spriteSizeToWidth(size);
        }

        /**
            @brief returns the height of the sprite.

            @return the width of the sprite.
        */
        inline const uint getHeight() const
        {
            return spriteSizeToHeight(size);
        }


        /**
         * @brief returns the color format of the sprite.
         *
         * @return the color format of the sprite, as an libnds enum value.
         */
        inline const SpriteColorFormat getColorFormat() const
        {
            return colorFormat;
        }


        /**
         * @brief sets a pointer to the graphics.
         *
         * sets a pointer to where the graphics of the sprite is located.
         *
         * @param gfxPointer a pointer to the graphics of the sprite.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setGfxPointer(u16* gfxPointer)
        {
            sassert(gfxPointer != NULL, "setGfxPointer() gfxPointer is NULL");

            releaseGFX();
            this->gfxPointer = gfxPointer;

            return *this;
        }


        /**
         * @brief returns a pointer to where the graphics of the sprite is located.
         *
         * @return a pointer to where the graphics of the sprite is located.
         */
        inline u16* getGfxPointer()
        {
            return gfxPointer;
        }


        /**
            @brief the constant version of getGfxPointer(), usable with constant objects.

            @return a const pointer to where the graphics of the sprite is located.
        */
        inline const u16* getGfxPointer() const
        {
            return gfxPointer;
        }



        /**
         * @brief sets a pointer to a AffineMatrix object.
         *
         * sets a pointer to a AffineMatrix object, which hold the data for affine transformation (rotating, scaling, ect).
         * seeing as there can be only a limited amount of AffineMatrix objects (32, a hardware limitation).
         * you can use this to let multiple sprites use the same AffineMatrix object.
         * if the sprite doesn't need affine transformation, you can set this to NULL.

            @param affineMatrixPointer a pointer to an AffineMatrix object, default is NULL.

            @return a reference to the sprite object, so you can chain multiple commands.
         */
        inline Sprite& setAffineMatrixPointer(AffineMatrix* affineMatrixPointer = NULL)
        {
            releaseAffineMatrix();

            this->affineMatrixPointer = affineMatrixPointer;

            return *this;
        }

        /**
         * @brief returns a pointer to a SpriteRotateScale object.
         *
         * this will return a pointer to the SpriteRotateScale object being used,
         * or 'NULL' if the sprite doesn't have a SpriteRotateScale object,
         * and thus isn't rotated or scaled.
         *
         * @return a pointer to where the graphics of the sprite is located.
         */
        inline AffineMatrix* getAffineMatrixPointer()
        {
            return affineMatrixPointer;
        }


        /**
            @brief the constant version of getRotateScalePointer(), usable with constant objects.

            @return a const pointer to where the graphics of the sprite is located.
        */
        inline const AffineMatrix* getAffineMatrixPointer() const
        {
            return affineMatrixPointer;
        }



        /**
         * @brief sets the x position of the sprite.
         *
         * @param xPos the x position of the sprite, default is 0.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setXPos(const int xPos = 0)
        {
            this->xPos = xPos;

            return *this;
        }

        /**
         * @brief sets the y position of the sprite.
         *
         * @param yPos the y position of the sprite, default is 0.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setYPos(const int yPos = 0)
        {
            this->yPos = yPos;

            return *this;
        }

        /**
         * @brief sets the x and y position of the sprite.
         *
         * @param xPos the x position of the sprite, default is 0.
         * @param yPos the y position of the sprite, default is 0.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setPosition(const int xPos = 0, const int yPos = 0)
        {
            this->xPos = xPos;
            this->yPos = yPos;

            return *this;
        }


        /**
         * @brief returns the x position of the sprite.
         *
         * @return the x position of the sprite.
         */
        inline const int getXPos() const
        {
            return xPos;
        }


        /**
         * @brief returns the y position of the sprite.
         *
         * @return the y position of the sprite.
         */
        inline const int getYPos() const
        {
            return yPos;
        }



        /**
         * @brief sets the sprites visibility.
         *
         * @param visible       true if the sprite should be visible, false if it should be hidden.
         *                      default is true.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setVisible(const bool visible = true)
        {
            this->visible = visible;

            return *this;
        }

        /**
         * @brief returns the sprites visibility.
         *
         * @return true if the sprite is visible, false otherwise.
         */
        inline const bool isVisible() const
        {
            return visible;
        }


        /**
         * @brief sets the priority of the sprite.
         *
         * sets the priority of the sprite.
         * the lower the priority, the later it is drawed, the more on top it is.
         *
         * @param priority the priority of the sprite (0-3), default is 0.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setPriority(const uint priority = 0)
        {
            sassert(priority < NUMBER_OF_PRIORITIES, "priority is out of bounds, needs to be between 0-3");
            this->priority = priority;

            return *this;
        }

        /**
         * @brief returns the priority of the sprite.
         *
         * @return the priority of the sprite (0-3).
         */
        inline const uint getPriority() const
        {
            return priority;
        }


        /**
         * @brief sets the sprites horizontal flip.
         *
         * @param flipHorizontal    true if the sprite should be horizontally flipped, false if it shouldn't be flipped.
         *                          default is false.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setHorizontalFlip(const bool flipHorizontal = false)
        {
            this->flipHorizontal = flipHorizontal;

            return *this;
        }

        /**
         * @brief returns the sprites horizontally flipped.
         *
         * @return true if the sprite is horizontally flipped, false otherwise.
         */
        inline const bool isHorizontalFlipped() const
        {
            return flipHorizontal;
        }

        /**
         * @brief sets the sprites vertical flip.
         *
         * @param flipVertical      true if the sprite should be vertically flipped, false if it shouldn't be flipped.
         *                          default is false.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setVerticalFlip(const bool flipVertical = false)
        {
            this->flipVertical = flipVertical;

            return *this;
        }

        /**
         * @brief returns the sprites vertically flipped.
         *
         * @return true if the sprite is vertically flipped, false otherwise.
         */
        inline const bool isVerticalFlipped() const
        {
            return flipVertical;
        }


        /**
         * @brief sets the sprites mosaic.
         *
         * @param mosaic      true if it should apply mosaic, false otherwise.
         *                    default is false.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setMosaic(const bool mosaic = false)
        {
            this->mosaic = mosaic;

            return *this;
        }

        /**
         * @brief returns if the sprite apply's mosaic is applied.
         *
         * @return true if a mosaic is applyed to the sprite, false otherwise.
         */
        inline const bool isMosaic() const
        {
            return mosaic;
        }


        /**
         * @brief sets if the sprites size should be doubled when the sprite is rotated.
         *
         * @param doubleSizeOnRotation      true if the sprites size should be doubled when the sprite is rotated, false otherwise.
         *                                  default is false.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setDoubleSizeOnRotation(const bool doubleSizeOnRotation = false)
        {
            this->doubleSizeOnRotation = doubleSizeOnRotation;

            return *this;
        }

        /**
         * @brief returns if the sprites size is doubled when the sprite is rotated.
         *
         * @return true if the sprites size is doubled when the sprite is rotated, false otherwise.
         */
        inline const bool getDoubleSizeOnRotation() const
        {
            return doubleSizeOnRotation;
        }


        /**
            @brief sets the sprites palette index or the alpha value.

            @param paletteIndexOrAlphaValue      the palette index if the sprite uses palettes,
                                                or the alpha value for bitmap sprites.

            @return a reference to the object, so you can chain multiple commands.
         */
        inline Sprite& setPaletteIndexOrAlphaValue(const int paletteIndexOrAlphaValue)
        {
            releasePalette();

            this->paletteIndexOrAlphaValue = paletteIndexOrAlphaValue;

            return *this;
        }

        /**
         * @brief   returns the sprites palette index or the alpha value.
         *
         * returns the sprites palette index if the sprite uses extended palletes
         * or the alpha value for bitmap sprites.
         *
         * @return true if the sprites size is doubled when the sprite is rotated, false otherwise.
         */
        inline const bool getPaletteIndexOrAlphaValue() const
        {
            return paletteIndexOrAlphaValue;
        }


        /**
            @brief returns the number of frames the sprite has.

            this will return 1 if you didn't provide the frame data with the constructor (if you used the second constructor).

            @return the number of framse the sprite has.
        */
        inline const uint getMaxNumberOfFrames() const
        {
            return maxNumberOfFrames;
        }


        /**
            @brief returns the current frame the sprite is on.

            this will return 1 if you didn't provide the frame data with the constructor (if you used the second constructor).

            @return the current frame the sprite is on.
        */
        inline const uint getCurrentFrame() const
        {
            return currentFrame;
        }

        /**
            @brief sets a new frame to the sprite, must be between 0 and (maxFrame-1).

            this will not work if you didn't provide the frame data with the constructor (if you used the second constructor).
            it will display an error message in debug mode.
            it will write the data directly, so there may be a glitch if you used it wrongly
            (during the drawing of the sprites to the screen).

            @param newFrame the new frame of the sprite, default is 0.

            @return a reference to the object, so you can chain multiple commands.
        */
        virtual Sprite& setFrame(const uint newFrame = 0);




        /**
         * @brief updates the sprite info to the oam 'shadow memory'.
         *
         * this will update the sprite info to the oam 'shadow memory' (the shadow memory is actually created by libnds).
         * you will still need to update the oam itself during vblank (will be done in the vblank() function of this library).
         *
         */
        inline virtual void update()
        {
            //adjust position when doubleSize is on.
            //flip sprite hor/ver when using matrix.

            oamSet(oamPointer,      //main graphics engine context
                    index,                      //oam index (0 to 127)
                    xPos, yPos,                 //x and y pixel location of the sprite
                    priority,                   //priority, lower renders last (on top)
                    paletteIndexOrAlphaValue,	//this is the palette index if multiple palettes or the alpha value if bmp sprite
                    size,                       //an libnds enum that tells the sprite size
                    colorFormat,                //an libnds enum that tells the sprite bitdebth (how many bits per pixel)
                    gfxPointer,                 //pointer to the loaded graphics
                    ((affineMatrixPointer == NULL) ? -1 : affineMatrixPointer->getIndex()),      //sprite rotation data
                    doubleSizeOnRotation,       //double the size when rotating?
                    !visible,			        //hide the sprite?
                    flipVertical,               //vflip
                    flipHorizontal,             //hflip
                    mosaic	                    //apply mosaic
                    );
        }


        /**
            @brief the deconstructor, it makes the sprite invisible and releases stuff.

            this deconstructor will hide the sprite, and it will also release the graphics,
            palettes and affine matrix it owns.
         */
        virtual ~Sprite();

};//end class Sprite


}//end namespace Sprites

}//end namespace DSOL

#endif // SPRITE_HPP_INC
