//including libnds
#include <nds.h>


#include "Sprite.hpp"


namespace DSOL
{

namespace Sprites
{

const uint spriteSizeToWidth(const SpriteSize size)
{
    switch (size)
    {
        case SpriteSize_8x8:
        case SpriteSize_8x16:
        case SpriteSize_8x32:
            return 8;

        case SpriteSize_16x8:
        case SpriteSize_16x16:
        case SpriteSize_16x32:
            return 16;

        case SpriteSize_32x8:
        case SpriteSize_32x16:
        case SpriteSize_32x32:
        case SpriteSize_32x64:
            return 32;

        case SpriteSize_64x32:
        case SpriteSize_64x64:
            return 64;
    }

    return 0;
}


const uint spriteSizeToHeight(const SpriteSize size)
{
    switch (size)
    {
        case SpriteSize_8x8:
        case SpriteSize_16x8:
        case SpriteSize_32x8:
            return 8;

        case SpriteSize_8x16:
        case SpriteSize_16x16:
        case SpriteSize_32x16:
            return 16;

        case SpriteSize_8x32:
        case SpriteSize_16x32:
        case SpriteSize_32x32:
        case SpriteSize_64x32:
            return 32;

        case SpriteSize_32x64:
        case SpriteSize_64x64:
            return 64;
    }

    return 0;
}






const uint gfxToByteSize(const SpriteSize size, const SpriteColorFormat colorFormat)
{
    register uint bytes = spriteSizeToPixel(size);

    if(colorFormat == SpriteColorFormat_16Color)
    {
        //2 pixels per byte
        bytes /= 2;
    }
    else if(colorFormat ==  SpriteColorFormat_Bmp)
    {
        //2 bytes per pixel
        bytes *= 2;
    }
    else
    {}//bytes stays the same

    return bytes;
}





void Sprite::releaseGFX()
{
    if(ownsGraphics)
    {
        sassert(gfxPointer != NULL, "sprite owns gfx but gfxPointer is NULL..");
        oamFreeGfx(oamPointer,  gfxPointer);
        gfxPointer = NULL;
        ownsGraphics = false;
    }
}


void Sprite::releasePalette()
{
    if(ownsPalette)
    {
        sassert(colorFormat != SpriteColorFormat_Bmp, "bitmap sprite owns palette (but doesn't uses a palette)...");
        Palette::freePalette(oamPointer, colorFormat == SpriteColorFormat_16Color, paletteIndexOrAlphaValue);
        paletteIndexOrAlphaValue = Palette::NUMBER_OF_PALETTES;
        ownsPalette = false;
    }
}


void Sprite::releaseAffineMatrix()
{
    if(ownsAffineMatrix)
    {
        sassert(affineMatrixPointer != NULL, "affineMatrixPointer is NULL while ownsAffineMatrix is set...");
        AffineMatrix::freeAffineMatrix(affineMatrixPointer->getIndex());
        deallocMem<AffineMatrix>(affineMatrixPointer);
        ownsAffineMatrix = false;
    }
}


//reconstruct constructor
Sprite::Sprite(const Sprite& source, const bool allocateNew, const bool takeOwnership, const uint unused /*= 0*/)
:
index(source.index), priority(source.priority), xPos(source.xPos), yPos(source.yPos), size(source.size),
visible(source.visible), flipVertical(source.flipVertical), flipHorizontal(source.flipHorizontal),
mosaic(source.mosaic), doubleSizeOnRotation(source.doubleSizeOnRotation),

dataPointer(source.dataPointer), dataLength(source.dataLength),
palettePointer(source.palettePointer), paletteLength(source.paletteLength),
maxNumberOfFrames(source.maxNumberOfFrames), currentFrame(source.currentFrame)
{
    sassert(source.gfxPointer != NULL, "error: source gfxPointer is NULL");

    //releasing gfx because sources sprite may need a different amount of vram
    releaseGFX();

    //releasing palette because it may not have the right size and oam
    releasePalette();
    colorFormat = source.colorFormat;
    oamPointer = source.oamPointer;

    //releasing the affine matrix because it may be for the wrong oam
    releaseAffineMatrix();



    if(allocateNew)
    {
        allocateGraphics();
        //copy data
        writeGraphics(reinterpret_cast<const u32*>(source.gfxPointer), gfxToByteSize(size, colorFormat));
    }
    else
    {
        gfxPointer = source.gfxPointer;
        //can only take ownership if the source has ownership
        if(takeOwnership && source.ownsGraphics)
        {
            ownsGraphics = true;
            source.ownsGraphics = false;
        }
    }


    //copy/allocate palette
    if(source.colorFormat != SpriteColorFormat_Bmp)
    {
        //we need a palette
        if(allocateNew)
        {
            allocatePalette();
            //copy palette
            Palette::copyPalette(source.colorFormat == SpriteColorFormat_16Color, source.paletteIndexOrAlphaValue,
                            source.oamPointer == &oamMain, paletteIndexOrAlphaValue, oamPointer == &oamMain);
        }
        else
        {
            paletteIndexOrAlphaValue = source.paletteIndexOrAlphaValue;
            //can only take ownership if the source has ownership
            if(takeOwnership && source.ownsPalette)
            {
                ownsPalette = true;
                source.ownsPalette = false;
            }
        }
    }


    //copy/allocate matrix
    if(source.affineMatrixPointer != NULL)
    {
        //we need an affine matrix
        if(allocateNew)
        {
            allocateAffineMatrix();

            //copy the affine matrix data from 1 place to the other (wont copy the index)
            *affineMatrixPointer = *(source.affineMatrixPointer);
        }
        else
        {
            affineMatrixPointer = source.affineMatrixPointer;
            //can only take ownership if the source has ownership
            if(takeOwnership && source.ownsAffineMatrix)
            {
                ownsAffineMatrix = true;
                source.ownsAffineMatrix = false;
            }
        }
    }
}




//first constructor
Sprite::Sprite( const uint index,
const u32* dataPointer, const uint dataLength,
const u16* palettePointer, const uint paletteLength,
const SpriteSize size, const SpriteColorFormat colorFormat,
const bool useMainOam /*= true*/,
const bool useAffineMatrix /*= false*/)
:
oamPointer(useMainOam ? &oamMain: &oamSub), index(index), priority(0),
gfxPointer(oamAllocateGfx(oamPointer, size, colorFormat)),
xPos(0), yPos(0), colorFormat(colorFormat), size(size),
visible(true), flipVertical(false), flipHorizontal(false),
mosaic(false), doubleSizeOnRotation(false),

ownsGraphics(true), ownsAffineMatrix(false),
dataPointer(dataPointer), dataLength(dataLength),
palettePointer(((colorFormat == SpriteColorFormat_Bmp) || (palettePointer == NULL)) ? NULL : palettePointer),
paletteLength(((colorFormat == SpriteColorFormat_Bmp) || (palettePointer == NULL)) ? 0 : paletteLength),
maxNumberOfFrames(gfxToFrameCount(dataLength, size, colorFormat)), currentFrame(0)
{
    sassert(index < NUMBER_OF_SPRITES, "index Sprite is not in 0-127");
    sassert(dataPointer != NULL, "error: Sprites dataPointer cannot be NULL");


    //copy gfx data
    copyGfxData(dataPointer, gfxPointer, gfxToByteSize(size, colorFormat));

    if ((colorFormat == SpriteColorFormat_Bmp) || (palettePointer == NULL))
    {
        paletteIndexOrAlphaValue = paletteLength;
        ownsPalette = false;
    }
    else
    {
        paletteIndexOrAlphaValue = Palette::reserveEmptyPalette(useMainOam, colorFormat == SpriteColorFormat_16Color);
        ownsPalette = true;

        //copy palette
        Palette::writeToPalette(useMainOam, colorFormat == SpriteColorFormat_16Color,
                                paletteIndexOrAlphaValue, palettePointer, paletteLength);
    }

    if(useAffineMatrix)
    {
        allocateAffineMatrix();
    }
    else
    {
        affineMatrixPointer = NULL;
    }
}


//copy constructor
Sprite::Sprite(const Sprite &source, const bool takeOwnership /*= true*/)
:
oamPointer(source.oamPointer), index(source.index), priority(source.priority), gfxPointer(source.gfxPointer),
xPos(source.xPos), yPos(source.yPos), colorFormat(source.colorFormat), size(source.size),
paletteIndexOrAlphaValue(source.paletteIndexOrAlphaValue), affineMatrixPointer(source.affineMatrixPointer),
visible(source.visible), flipVertical(source.flipVertical), flipHorizontal(source.flipHorizontal),
mosaic(source.mosaic), doubleSizeOnRotation(source.doubleSizeOnRotation),

ownsGraphics(takeOwnership && source.ownsGraphics), ownsPalette(takeOwnership && source.ownsPalette),
ownsAffineMatrix(takeOwnership && source.ownsAffineMatrix),
dataPointer(source.dataPointer), dataLength(source.dataLength),
palettePointer(source.palettePointer), paletteLength(source.paletteLength),
maxNumberOfFrames(source.maxNumberOfFrames), currentFrame(source.currentFrame)
{
    //doesn't have to release vram ect.
    sassert(source.gfxPointer != NULL, "error: source gfxPointer is NULL");

    source.ownsGraphics     = (!takeOwnership) && source.ownsGraphics;
    source.ownsPalette      = (!takeOwnership) && source.ownsPalette;
    source.ownsAffineMatrix = (!takeOwnership) && source.ownsAffineMatrix;
}



Sprite& Sprite::allocateGraphics()
{
    if(!ownsGraphics)
    {
        gfxPointer = oamAllocateGfx(oamPointer, size, colorFormat);
        ownsGraphics = true;
    }

    return *this;
}



Sprite& Sprite::allocatePalette()
{
    if(!ownsPalette && colorFormat != SpriteColorFormat_Bmp)
    {
        paletteIndexOrAlphaValue = Palette::reserveEmptyPalette((oamPointer == &oamMain), colorFormat == SpriteColorFormat_16Color);
        ownsPalette = true;
    }

    return *this;
}


Sprite& Sprite::allocateAffineMatrix()
{
    if(!ownsAffineMatrix)
    {
        affineMatrixPointer = allocMem<AffineMatrix>(AffineMatrix(
                                AffineMatrix::reserveEmptyAffineMatrix(oamPointer == &oamMain),
                                                                        (oamPointer == &oamMain)));

        //asserting will already be done by allocMem() and reserveEmptyAffineMatrix()
        //so no need to do it here

        ownsAffineMatrix = (affineMatrixPointer != NULL);
    }

    return *this;
}



/*virtual*/ Sprite& Sprite::setFrame(const uint newFrame /*= 0*/)
{
    sassert(dataPointer != NULL, "error: sprite doesn't have frame data");
    sassert(newFrame < maxNumberOfFrames, "error: newFrame is out of bounds");
    sassert(gfxPointer != NULL, "gfxPointer of sprite is NULL");

    register uint byteSize = gfxToByteSize(size, colorFormat);
    //copy gfx data
    copyGfxData(dataPointer + ((newFrame * byteSize) / sizeof(*dataPointer)), gfxPointer, byteSize);

    currentFrame = newFrame;

    return *this;
}



/*virtual*/ Sprite::~Sprite()
{
    oamClear(oamPointer, index, 1); //disables only this sprite

    releaseAffineMatrix();
    releasePalette();
    releaseGFX();
}


}// end namespace Sprites

}//end namespace DSOL
